import tensordict
from torch.utils.data import DataLoader
import torch

from crtl.components.dual_encoder.pipeline.embedder import Embedder

from crtl.components.dual_encoder.pipeline.negative_sampler import NegativeSampler
from crtl.components.dual_encoder.pipeline.utils import convert_size


class DataLoaderCreator:
    def __init__(
        self,
        number_of_random_negatives: int,
        number_of_hard_negatives: int,
        device: torch.device,
        batch_size: int,
    ):
        self.negative_sampler = NegativeSampler(
            number_of_random_negatives, number_of_hard_negatives, device
        )
        self.batch_size = batch_size
        self.device = device

    def __call__(
        self,
        gold_candidate_indices: torch.Tensor,
        mention_encodings: tensordict.TensorDictBase,
        candidate_encodings: tensordict.TensorDictBase,
        mention_embeddings: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ) -> DataLoader[tensordict.TensorDict]:
        dataset = self.get_dataset(
            gold_candidate_indices,
            mention_encodings,
            candidate_encodings,
            mention_embeddings,
            candidate_embeddings,
        ).to(self.device)
        return DataLoader(
            dataset, batch_size=self.batch_size, shuffle=True, collate_fn=lambda x: x
        )

    def get_dataset(
        self,
        gold_candidate_indices: torch.Tensor,
        mention_encodings: tensordict.TensorDictBase,
        candidate_encodings: tensordict.TensorDictBase,
        mention_embeddings: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ) -> tensordict.TensorDict:
        positive_candidate_encodings = candidate_encodings.gather(
            0, gold_candidate_indices
        )
        negative_candidate_encodings = self.get_negative_candidate_encodings(
            gold_candidate_indices,
            candidate_encodings,
            mention_embeddings,
            candidate_embeddings,
        )
        return tensordict.TensorDict(
            {
                "mention_encodings": mention_encodings,
                "positive_candidate_encodings": positive_candidate_encodings,
                "negative_candidate_encodings": negative_candidate_encodings,
            },
            batch_size=len(mention_encodings),
        )

    def get_negative_candidate_encodings(
        self,
        gold_candidate_indices,
        candidate_encodings,
        mention_embeddings,
        candidate_embeddings,
    ):
        negative_candidate_ids = self.negative_sampler.get_negatives(
            gold_candidate_indices, mention_embeddings, candidate_embeddings
        ).flatten()  # num_mentions x (num_random_negatives + num_hard_negatives)
        return candidate_encodings.gather(0, negative_candidate_ids).reshape(
            gold_candidate_indices.shape[0], -1
        )


def test_dataset_creator():
    tokenizer = AutoTokenizer.from_pretrained("sentence-transformers/all-MiniLM-L6-v2")
    mention_encoder = Embedder()
    candidate_encoder = Embedder()
    candidate_texts = [
        "We are going to the theaters to see a movie.",
        "I can't wait to tell my husband I made the volleyball team!",
        "I can barely hit the ball.",
        "I was recently diagnosed with cancer.",
        "You can't let your illness define you. ",
    ]
    mention_texts = [
        "My kid just made the team.",
        "Receiving a diagnosis is always hard.",
    ]
    dataset_creator = DatasetCreator(
        mention_encodings=mention_encoder.tokenize(mention_texts),
        candidate_encodings=candidate_encoder.tokenize(candidate_texts),
        gold_candidate_indices=torch.tensor([1, 3]),
        number_of_random_negatives=1,
        number_of_hard_negatives=1,
    )
    dataset = dataset_creator(mention_encoder, candidate_encoder)
    print(dataset[0])
    print(dataset[1])
