from typing import List, Iterable, Mapping, NamedTuple, Tuple, Callable, Optional, Dict


import torch
import torch.nn.functional as F

from transformers import AutoTokenizer, PreTrainedTokenizerBase, AutoModel, BertModel
from torch import nn
from tensordict import TensorDict, TensorDictBase


class EmbedderModule(nn.Module):
    def __init__(self, bert_model: BertModel):
        super().__init__()
        self.model = bert_model

    def _forward(
        self, input_ids: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)

        # Perform pooling
        sentence_embeddings = self.mean_pooling(outputs, attention_mask)

        # Normalize embeddings
        sentence_embeddings = F.normalize(sentence_embeddings, p=2, dim=1)

        return sentence_embeddings.squeeze()

    def forward(self, torch_dict: TensorDictBase):
        return self._forward(
            torch_dict.get("input_ids"), torch_dict.get("attention_mask")
        )

    def mean_pooling(
        self, model_output: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        """
        mean_pooling

        Applies a mean pooling to the last hidden layer of the underlying model using an attention mask

        ------
        PARAMS:

        model_output: BaseModelOutputWithPoolingAndCrossAttentions
            The output of the model after being fed input id's

        attention_mask: Tensor
            The attention mask for the token embeddings
        """

        # First element of model_output contains all token embeddings
        token_embeddings = model_output[0]

        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )


class Embedder(NamedTuple):
    tokenizer: PreTrainedTokenizerBase
    embedder_module: EmbedderModule

    @classmethod
    def from_pretrained(
        cls, pretrained_hf_bert_path: str = "sentence-transformers/all-MiniLM-L6-v2"
    ):
        return cls(
            AutoTokenizer.from_pretrained(pretrained_hf_bert_path),
            EmbedderModule(AutoModel.from_pretrained(pretrained_hf_bert_path)),
        )
