from typing import Callable, Literal
import torch

from crtl.components.dual_encoder.pipeline.similarity import DotSimilarity


class CrossEntropyScoreLoss(torch.nn.Module):
    """
    CrossEntropyLoss

    Loss over dot product distance scores.

    pool_negatives:
    """

    def __init__(self, batch_size):
        super(CrossEntropyScoreLoss, self).__init__()
        self.ce_loss = torch.nn.CrossEntropyLoss()

    def forward(
        self,
        mention_embedding: torch.Tensor,  # batch_size x embedding_dim
        positive_embedding: torch.Tensor,  # batch_size x embedding_dim
        negative_embeddings: torch.Tensor,  # batch_size x (num_rand_negatives + num_hard_negatives) x embedding_dim
    ) -> torch.Tensor:
        """
        forward

        Returns the cross entropy loss of the similarity scores between the mention
        embedding and the postive and negative candidate embeddings, using the positive candidate
        as the correct label.

        -------
        PARAMS:

        predictions: Tensor
            The embedding predicted by the mention encoder model from a mention
        positives: Tensor
            The embedding created by the concept encoder model from the postive concept corresponding to the mention
        negatives: Tensor
            A list of embeddings created by the concept encoder model from the list of negative concepts corresponding to the mention
        """
        candidate_embeddings = torch.cat(
            [positive_embedding.unsqueeze(1), negative_embeddings], dim=1
        )  # batch_size x (1 + (num_rand_negatives + num_hard_negatives)) x embedding_dim
        score = torch.bmm(
            candidate_embeddings,
            mention_embedding.unsqueeze(-1),  # batch_size x embeding_dimension x 1
        ).squeeze(
            -1
        )  # batch_size x (1 + (num_rand_negatives + num_hard_negatives))
        return self.ce_loss(
            score, torch.zeros(score.size(0), dtype=torch.uint8, device=score.device)
        )


class MarginMSELoss(torch.nn.Module):
    """
    MarginMSELoss

    A custom torch loss function that computes margin MSE loss (similar to TripletMarginWithDistanceLoss with a dot product distance but with the ability to support multiple negatives per positive)

    pool_negatives:
        if true, weights averages of all negative scores per example into the weight of one positive example
    """

    def __init__(
        self,
        pool_negatives: bool = True,
        similarity_calculator: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = DotSimilarity(),
        margin: float = 1.0,
        reduction_strategy: Literal["mean", "sum"] = "mean",
    ):
        super(MarginMSELoss, self).__init__()
        self.pool_negatives = pool_negatives
        self.similarity_calculator = similarity_calculator
        self.margin = 1.0
        self.reduction_strategy = reduction_strategy
        # self.mse = torch.nn.MSELoss()

    def reduce(self, scores: torch.Tensor) -> torch.Tensor:
        if self.reduction_strategy == "sum":
            return torch.sum(scores)
        elif self.reduction_strategy == "mean":
            return torch.mean(scores)
        else:
            raise ValueError(
                f"self.reduction_strategy should be one of 'sum' or 'mean', not {self.reduction_strategy}"
            )

    def forward(
        self,
        prediction_embeddings: torch.Tensor,  # num_mentions x num_negatives x embedding_dim
        positive_embedding: torch.Tensor,  # num_mentions x embedding_dim
        negative_embeddings: torch.Tensor,  # num_mentions x num_negatives x embedding_dim
    ) -> torch.Tensor:
        """
        forward

        Returns the marginMSE loss function of the encoders over a list of encoder predictions with positive and negative examples.

        More details on marginMSE loss here: https://arxiv.org/pdf/2010.02666.pdf

        -------
        PARAMS:

        predictions: Tensor
            The embedding predicted by the mention encoder model from a mention
        positives: Tensor
            The embedding created by the concept encoder model from the postive concept corresponding to the mention
        negatives: Tensor
            A list of embeddings created by the concept encoder model from the list of negative concepts corresponding to the mention
        """
        candidate_embeddings = torch.cat(
            [positive_embedding.unsqueeze(1), negative_embeddings], dim=1
        )
        similarities = self.similarity_calculator(
            prediction_embeddings, candidate_embeddings
        )  # num_mentions x num_candidates_per_mention

        positive_scores = similarities[:, 0].squeeze(1)  # num_mentions
        negative_scores = similarities[:, 1:]  # num_mentions x num_negatives
        if self.pool_negatives:
            negative_scores = torch.mean(negative_scores, dim=1)  # num_mentions

        else:
            negative_scores = torch.sum(negative_scores, dim=1)  # num_mentions

        actual = torch.max(
            positive_scores - negative_scores + self.margin,
            torch.zeros(prediction_embeddings.shape[0]),
        )

        return self.reduce(actual)


class CrossEntropyScoreLoss2(torch.nn.Module):
    """
    CrossEntropyLoss

    Loss over dot product distance scores.

    pool_negatives:
    """

    def __init__(
        self,
        similarity_calculator: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = DotSimilarity(),
    ):
        super(CrossEntropyScoreLoss2, self).__init__()
        self.ce_loss = torch.nn.CrossEntropyLoss()
        self.similarity_calculator = similarity_calculator

    def forward(
        self,
        mention_embeddings: torch.Tensor,  # batch_size x embedding_dim
        positive_embeddings: torch.Tensor,  # batch_size x embedding_dim
        negative_embeddings: torch.Tensor,  # batch_size x (num_rand_negatives + num_hard_negatives) x embedding_dim
    ) -> torch.Tensor:
        """
        forward

        Returns the cross entropy loss of the similarity scores between the mention
        embedding and the postive and negative candidate embeddings, using the positive candidate
        as the correct label.

        -------
        PARAMS:

        predictions: Tensor
            The embedding predicted by the mention encoder model from a mention
        positives: Tensor
            The embedding created by the concept encoder model from the postive concept corresponding to the mention
        negatives: Tensor
            A list of embeddings created by the concept encoder model from the list of negative concepts corresponding to the mention
        """
        candidate_embeddings = torch.cat(
            [positive_embeddings.unsqueeze(1), negative_embeddings], dim=1
        )  # num_mentions x (1 + (num_rand_negatives + num_hard_negatives)) x embedding_dim
        score = self.similarity_calculator(
            mention_embeddings, candidate_embeddings
        )  # num_mentions x (1 + (num_rand_negatives + num_hard_negatives))
        return self.ce_loss(
            score, torch.zeros(score.size(0), dtype=torch.uint8, device=score.device)
        )
