import itertools
from typing import Iterable, List, Tuple
import torch
from crtl.components.dual_encoder.pipeline.preprocessor import (
    EncodedMentions,
    LKBPreprocessor,
    MentionPreprocessor,
    PreproccessedDataSet,
    Preprocessor,
)
from crtl.datasets.medmentions.parse_medmentions import (
    MedMentionCorpus,
    MedMentionDocument,
    Mention,
)

from crtl.lkb.lkb import LexicalKnowledgeBase
from transformers import PreTrainedTokenizerBase


class MedmentionMentionEncoder:
    def __init__(self, tokenizer: PreTrainedTokenizerBase):
        self._tokenizer = tokenizer

    def __call__(
        self, docs: Iterable[MedMentionDocument], lkb: LexicalKnowledgeBase
    ) -> EncodedMentions:
        doc_encodings = self._tokenizer([doc.full_text for doc in docs])
        docs_with_filtered_mentions = [
            (doc, self._get_filtered_mentions(doc.mentions, lkb)) for doc in docs
        ]
        mention_indices = [
            [
                self._get_mention_start_and_end_indices(doc.full_text, mention)
                for mention in mentions
            ]
            for doc, mentions in docs_with_filtered_mentions
        ]
        gold_candidate_ids = [
            [mention.entity_id for mention in filtered_mentions]
            for doc, filtered_mentions in docs_with_filtered_mentions
        ]
        return EncodedMentions(doc_encodings, mention_indices, gold_candidate_ids)

    def _get_filtered_mentions(
        self, mentions: Iterable[Mention], lkb: LexicalKnowledgeBase
    ) -> Iterable[Mention]:
        return [
            mention for mention in mentions if mention.entity_id in lkb.cui_to_concept
        ]

    def _get_mention_start_and_end_indices(
        self, full_text: str, mention: Mention
    ) -> Tuple[int, int]:
        mention_start_index = mention.start_index
        mention_end_index = mention.end_index - 1
        if (
            full_text[mention.start_index] == "\u200e"
            or full_text[mention.start_index] == "\u200f"
        ):
            mention_start_index += 1
        if (
            full_text[mention_end_index] == "\u200e"
            or full_text[mention_end_index] == "\u200f"
        ):
            mention_end_index -= 1
        return (
            mention_start_index,
            mention_end_index,
        )


class MedmentionsPreprocessor:
    def __init__(
        self,
        mention_tokenizer: PreTrainedTokenizerBase,
        length: int,
        lkb_preprocessor: LKBPreprocessor,
        eval_split: float,
    ):
        self._medmention_encoder = MedmentionMentionEncoder(mention_tokenizer)
        self._eval_split = eval_split
        self._preprocessor = Preprocessor(
            MentionPreprocessor(mention_tokenizer, length), lkb_preprocessor
        )

    def split(self, docs: List[MedMentionDocument]):
        num_training = len(docs) - int(len(docs) * self._eval_split)
        return docs[:num_training], docs[num_training:]

    def __call__(
        self, medmentions_corpus: MedMentionCorpus, lkb: LexicalKnowledgeBase
    ) -> PreproccessedDataSet:
        train_docs, val_docs = self.split(medmentions_corpus.documents)
        return self._preprocessor(
            self._medmention_encoder(train_docs, lkb),
            self._medmention_encoder(val_docs, lkb),
            lkb,
        )
