import random
from typing import Callable, Iterable, List, Sequence
from annotated_types import T
import torch

from crtl.components.dual_encoder.pipeline.similarity import get_dot_similarity
from crtl.components.dual_encoder.pipeline.utils import convert_size


class NegativeSampler:
    def __init__(
        self,
        number_random_negatives,
        number_hard_negatives,
        device,
        similarity_calculator: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = get_dot_similarity,
    ):
        self.number_random_negatives = number_random_negatives
        self.number_hard_negatives = number_hard_negatives
        self.similarity_calculator = similarity_calculator
        self.device = device

    def get_negatives(
        self,
        gold_candidate_indices: torch.Tensor,
        mention_embeddings: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ):
        with torch.no_grad():
            num_mentions = mention_embeddings.shape[0]
            sorted_similarities = self.get_sorted_similarities(
                gold_candidate_indices, mention_embeddings, candidate_embeddings
            )
            hard_negatives = sorted_similarities[
                :, : self.number_hard_negatives
            ]  # num_mentions x num_hard_negatives
            random_negatives = self.get_random_negatives(
                num_mentions, sorted_similarities[:, self.number_hard_negatives : -1]
            )
            return torch.cat(
                [
                    random_negatives,
                    hard_negatives,
                ],
                dim=1,
            ).to(self.device)

    def get_mask(self, gold_candidate_indices, num_mentions, num_candidates):
        mask = torch.zeros((num_mentions, num_candidates)).to(self.device)
        mask[torch.arange(mask.size(0)), gold_candidate_indices] = -torch.inf
        return mask

    def get_sorted_similarities(
        self, gold_candidate_indices, mention_embeddings, candidate_embeddings
    ):
        similarities = self.get_masked_similarity(
            gold_candidate_indices, mention_embeddings, candidate_embeddings
        )
        sorted_similarities = torch.argsort(similarities, dim=1, descending=True)
        return sorted_similarities

    def get_masked_similarity(
        self, gold_candidate_indices, mention_embeddings, candidate_embeddings
    ):
        num_mentions = mention_embeddings.shape[0]
        num_candidates = candidate_embeddings.shape[0]
        mask = self.get_mask(gold_candidate_indices, num_mentions, num_candidates).to(
            torch.device("cpu")
        )
        similarity = (
            self.similarity_calculator(mention_embeddings, candidate_embeddings) + mask
        )
        return similarity

    def get_random_negatives(self, num_mentions, random_candidates: torch.Tensor):
        weights = torch.ones(random_candidates.shape)
        # random_selections = torch.randint(0,random_candidates.shape[-1],(mention_embeddings.shape[0],self.number_random_negatives)).to(self.device)
        random_selections = torch.multinomial(
            weights, num_samples=self.number_random_negatives, replacement=False
        )
        return random_candidates[
            torch.arange(num_mentions).repeat((self.number_random_negatives, 1)).T,
            random_selections,
        ]


class NegativeSamplerAlt:
    def __init__(
        self,
        max_number_candidates,
        max_number_hard_negatives,
        device,
        similarity_calculator: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = get_dot_similarity,
    ):
        self.max_number_candidates = max_number_candidates
        self.max_number_hard_negatives = max_number_hard_negatives
        self.similarity_calculator = similarity_calculator
        self.device = device

    def get_hard_negative_indices(
        self,
        gold_candidate_indices: torch.Tensor,
        mention_embeddings: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ):
        similarities = self.similarity_calculator(
            mention_embeddings, candidate_embeddings
        )  # num_mentions x num_candidates
        ranked_candidate_indices = torch.argsort(
            similarities, dim=1, descending=True
        )  # num_mentions x num_candidates
        gold_rankings = (
            ranked_candidate_indices == gold_candidate_indices
        ).nonzero()  # num_mentions
        candidates_sorted_by_similarity_indices = torch.arange(
            0, len(candidate_embeddings)
        ).repeat(
            len(mention_embeddings)
        )  # num_mentions x num_candidates
        greater_than_gold_rank = (
            candidates_sorted_by_similarity_indices < gold_rankings.unsqueeze(-1)
        )
        fewer_than_max = (
            candidates_sorted_by_similarity_indices < self.max_number_hard_negatives
        )
        is_hard_candidate = fewer_than_max and greater_than_gold_rank
        num_hard_negatives_per_mention = torch.sum(is_hard_candidate, dim=1)
        candidate_ranks = (
            is_hard_candidate.nonzero()
        )  # hard_candidates x 2 (1:mention_index,2:candidate_rank)
        hard_candidate_indices = ranked_candidate_indices[
            candidate_ranks[:, 0], candidate_ranks[:, 1]
        ]
        return (
            torch.cat([candidate_ranks[:, 0], hard_candidate_indices], dim=1),
            num_hard_negatives_per_mention,
        )  ##hard_candidates x 2 (1:mention_index, 2:candidate_index), num_mentions

    def get_random_indices(
        self,
        num_candidates: int,
        gold_candidate_indices,  # num_mentions
        hard_candidate_indices,  # hard_candidates x 2 (1:mention_index,2:candidate_index)
        num_hard_negatives_per_mention,  # num_mentions x max_num_hard_candidates
    ):

        weight_matrix = torch.ones(
            len(gold_candidate_indices), num_candidates
        )  # num_mentions x num_candidates
        weight_matrix[torch.arange(gold_candidate_indices), gold_candidate_indices] = 0
        weight_matrix[hard_candidate_indices[:, 0], hard_candidate_indices[:, 1]] = 0
        random_candidates = torch.multinomial(
            weight_matrix, num_samples=self.max_number_candidates, replacement=False
        )  # num_mentions x max_num_candidates
        random_candidates_needed_per_mention = (
            self.max_number_candidates - num_hard_negatives_per_mention
        )
        random_candidate_indices = (
            torch.arange(0, num_candidates).repeat(len(gold_candidate_indices))
            < random_candidates_needed_per_mention.unsqueeze(-1)
        ).nonzero()
        hard_candidate_indices = random_candidates[
            random_candidate_indices[:, 0], random_candidate_indices[:, 1]
        ]
        return torch.cat(
            [random_candidate_indices[:0], random_candidate_indices], dim=1
        )  ##random_candidates x 2 (1:mention_index, 2:candidate_index)

    def get_negatives(
        self,
        gold_candidate_indices: torch.Tensor,
        mention_embeddings: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ):
        with torch.no_grad():
            (
                hard_candidate_indices,
                num_hard_negatives_per_mention,
            ) = self.get_hard_negative_indices(
                gold_candidate_indices, mention_embeddings, candidate_embeddings
            )
            random_candidate_indices = self.get_random_indices(
                len(candidate_embeddings),
                gold_candidate_indices,
                hard_candidate_indices,
                num_hard_negatives_per_mention,
            )
            flat_candidate_indices = torch.cat(
                [hard_candidate_indices, random_candidate_indices]
            )
            mention_indices_in_flat_candidate_indices = torch.argsort(
                flat_candidate_indices[:, 0]
            )
            candidate_indices = torch.zeros(
                len(mention_embeddings), self.max_number_candidates
            )
            candidate_indices[
                mention_indices_in_flat_candidate_indices,
                torch.arange(0, self.max_number_candidates).repeat(
                    len(mention_embeddings)
                ),
            ] = flat_candidate_indices[mention_indices_in_flat_candidate_indices, 1]
            return candidate_indices  # num_mentions x MAX_NUMBER_CANDIDATES


# class NegativeSampler:
#     def __init__(self,number_random_negatives,number_hard_negatives,device):
#         self.number_random_negatives = number_random_negatives
#         self.number_hard_negatives = number_hard_negatives
#         self.device = device
#     def get_negatives(self,gold_candidate_indices:torch.Tensor,mention_embeddings:torch.Tensor,candidate_embeddings:torch.Tensor):
#         random_negatives = self.get_random_negatives(gold_candidate_indices,len(candidate_embeddings))

#         hard_negatives = self.get_hard_negatives(
#             gold_candidate_indices,mention_embeddings, candidate_embeddings, random_negatives
#         )

#         return torch.cat(
#             [
#                 random_negatives,
#                 hard_negatives,
#             ],
#             dim=1
#         )
#     def _random_sample_with_exclusions(self,pop:Sequence[T],num:int,excluded_value:T)->List[T]:
#         match:set[T] = set()
#         num_matches = 0
#         while num_matches < num:
#             x = random.sample(pop,1)[0]
#             if x != excluded_value and x not in match:
#                 num_matches+=1
#                 match.add(x)
#         return list(match)


#     def get_random_negatives(self,gold_candidate_indices:torch.Tensor,num_candidates:int):
#         gold_candidate_indices_list:list = gold_candidate_indices.tolist()
#         all_possible_mention_indices = range(num_candidates)
#         random_negative_ids = []
#         for i in progressbar(gold_candidate_indices,len(gold_candidate_indices_list)):
#             sample = self._random_sample_with_exclusions(all_possible_mention_indices,num=self.number_random_negatives,excluded_value=i)
#             random_negative_ids.append(
#                 sample
#             )

#         return torch.tensor(random_negative_ids).to(self.device)  # no_mentions x no_random_negatives

#     def get_hard_negatives(
#         self,
#         gold_candidate_indices:torch.Tensor,
#         mention_embeddings: torch.Tensor,
#         candidate_embeddings: torch.Tensor,
#         random_negatives: torch.Tensor,
# ) -> torch.Tensor:
#         mask = self.get_candidate_mask(gold_candidate_indices,random_negatives,len(mention_embeddings),len(candidate_embeddings))
#         dot_similarity = torch.matmul(mention_embeddings,candidate_embeddings.T) + mask# no_mentions x no_candidates

#         # cosine_similarity = (
#         #     torch.cosine_similarity(
#         #         mention_embeddings[:, None, :], candidate_embeddings[None, :, :], dim=-1
#         #     )
#         #     + mask
#         # )
#         top_indices = dot_similarity.topk(
#             self.number_hard_negatives, dim=1
#         ).indices  # no_mentions x no_hard_negatives
#         return top_indices
#     def get_candidate_mask(self,gold_candidate_indices, random_negatives: torch.Tensor,num_mention_encodings,num_candidate_encodings):
#         # create a mask to apply to the cosine similarity so we don't include gold or random negatives
#         mask = torch.zeros((num_mention_encodings, num_candidate_encodings))
#         mask[torch.arange(mask.size(0)), gold_candidate_indices] = -torch.inf
#         mask[
#             torch.arange(mask.size(0)).repeat((self.number_random_negatives, 1)).T,
#             random_negatives,
#         ] = -torch.inf
#         return mask.to(self.device)  # no_mentions x no_candidates
