import itertools
import pathlib
from typing import List, Mapping, NamedTuple, Tuple, Union
import tensordict
import torch

from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from transformers import PreTrainedTokenizerBase, BatchEncoding
import re


class PreproccesedDatasetPartition(NamedTuple):
    mention_encodings: tensordict.TensorDictBase
    gold_candidate_ids: torch.Tensor

    def to_device(self, device: torch.device) -> "PreproccesedDatasetPartition":
        return PreproccesedDatasetPartition(
            self.mention_encodings.to(device), self.gold_candidate_ids.to(device)
        )


class PreproccessedDataSet(NamedTuple):
    training_dataset_partition: PreproccesedDatasetPartition
    validation_dataset_partition: PreproccesedDatasetPartition
    candidate_encodings: tensordict.TensorDictBase

    def to_device(self, device) -> "PreproccessedDataSet":
        return PreproccessedDataSet(
            self.training_dataset_partition.to_device(device),
            self.validation_dataset_partition.to_device(device),
            self.candidate_encodings.to(device),
        )


def convert_path(path: Union[str, pathlib.Path]) -> pathlib.Path:
    if type(path) == str:
        pathlib_path = pathlib.Path(path)
    elif type(path) == pathlib.Path:
        pathlib_path = path
    else:
        raise TypeError("path must be one of 'str' or 'pathlib.Path'")
    return pathlib_path


# def load_torch_mapping_from_disk(
#     path:Union[str,pathlib.Path]
# )->Mapping[str, torch.Tensor]:
#     pathlib_path = convert_path(path)
#     [file_path for file_path in pathlib_path.iterdir()]


# def send_torch_mapping_to_disk(
#     tensor_mapping: Mapping[str, torch.Tensor],path:Union[str,pathlib.Path]
# )->None:
#    pathlib_path = convert_path(path)
#    for key,value in tensor_mapping:
#        torch.save(value,pathlib_path / key)


def send_torch_mapping_to_device(
    tensor_mapping: Mapping[str, torch.Tensor], device: torch.device
):
    for value in tensor_mapping.values():
        value.to(device)


LKBIndex = Mapping[str, int]


class LKBPreprocessor:
    def __init__(self, tokenizer: PreTrainedTokenizerBase, length: int):
        self._tokenizer = tokenizer
        self._length = length

    def get_concepts(self, lkb: LexicalKnowledgeBase) -> list[Concept]:
        return list(lkb.cui_to_concept.values())

    def get_concept_index(self, concepts: list[Concept]) -> LKBIndex:
        return {str(concept.concept_id): i for i, concept in enumerate(concepts)}

    def get_concept_encodings(
        self, concepts: list[Concept]
    ) -> Mapping[str, torch.Tensor]:
        encoding = self._tokenizer(
            [self.get_concept_representation(concept) for concept in concepts],
            return_tensors="pt",
            max_length=self._length,
            padding="max_length",
            truncation=True,
            # return_length=True
        )
        return encoding

    def get_concept_representation(self, concept: Concept):
        terms = set(
            tuple(re.split(",|-| ", term)) for term in concept.terms if len(term) < 100
        )
        terms = [
            " ".join(term_segment for term_segment in term if term_segment != "")
            for term in terms
        ]
        concept_repr = "|".join(terms)
        concept_repr = concept_repr if concept_repr else concept.canonical_name
        # concept_repr =  f"{concept.canonical_name}{f':{concept.definition}' if concept.definition else ''}"
        print(concept_repr)
        return concept_repr

    def __call__(
        self, lkb: LexicalKnowledgeBase
    ) -> Tuple[Mapping[str, torch.Tensor], LKBIndex, int]:
        concepts = self.get_concepts(lkb)
        return (
            self.get_concept_encodings(concepts),
            self.get_concept_index(concepts),
            len(concepts),
        )


class EncodedMentions(NamedTuple):
    doc_encodings: BatchEncoding
    mention_indices: List[List[Tuple[int, int]]]
    gold_concept_ids: List[List[str]]


class Windower:
    def __init__(self, length):
        self._length = length

    def _get_context(
        self, tokenized_doc: list[str], span_start_index: int, span_end_index: int
    ) -> list[str]:
        number_of_tokens_in_span = span_end_index - span_start_index
        number_of_tokens_remaining = self._length - (
            number_of_tokens_in_span * 2 + 3
        )  # 3 special tokens and 2 occurences of the span
        left_context_length = number_of_tokens_remaining // 2
        right_context_length = left_context_length + number_of_tokens_remaining % 2
        (
            context_start_index,
            context_end_index,
        ) = self._get_modified_context_start_and_end_indices(
            len(tokenized_doc),
            span_start_index,
            span_end_index,
            left_context_length,
            right_context_length,
        )
        return tokenized_doc[context_start_index:context_end_index]

    def _get_modified_context_start_and_end_indices(
        self,
        tokenized_doc_len: int,
        start_token_index: int,
        end_token_index: int,
        left_context_length: int,
        right_context_length: int,
    ):
        left_oob = start_token_index - left_context_length < 0  # left out-of-bounds
        right_oob = (
            end_token_index + right_context_length > tokenized_doc_len
        )  # right-out-of-bounds
        if not (left_oob or right_oob):  # both fit within the document, do nothing
            context_start_index = start_token_index - left_context_length
            context_end_index = end_token_index + right_context_length
        elif (
            left_oob and not right_oob
        ):  # 2 only left is out of bounds, give difference to the right context length
            left_oob_amount = -(start_token_index - left_context_length)
            context_start_index = 0
            context_end_index = min(
                end_token_index + right_context_length + left_oob_amount,
                tokenized_doc_len,
            )
        elif (
            right_oob and not left_oob
        ):  # 3 only right is out of bounds, give difference to the right context length
            right_oob_amount = (
                end_token_index + right_context_length - tokenized_doc_len
            )
            context_start_index = max(
                start_token_index - left_context_length - right_oob_amount, 0
            )
            context_end_index = tokenized_doc_len
        else:  # 4 both are out of bounds, return whole list
            context_start_index = 0
            context_end_index = tokenized_doc_len
        return context_start_index, context_end_index

    def _get_spans_and_context(
        self,
        batch_encoding: BatchEncoding,
        mention_indices: List[List[Tuple[int, int]]],
    ) -> Tuple[List[List[str]], List[List[str]]]:
        batch_span_tokens = []
        batch_context_tokens = []
        for doc_index, doc_mention_indices in enumerate(mention_indices):
            for mention_start_index, mention_end_index in doc_mention_indices:
                document_encoding = batch_encoding.tokens(doc_index)[1:-1]
                start_token_index = (
                    batch_encoding.char_to_token(doc_index, mention_start_index) - 1
                )
                end_token_index = batch_encoding.char_to_token(
                    doc_index, mention_end_index
                )
                context_tokens = self._get_context(
                    document_encoding, start_token_index, end_token_index
                )
                span_tokens = document_encoding[start_token_index:end_token_index]
                batch_span_tokens.append(span_tokens)
                batch_context_tokens.append(context_tokens)
        return batch_span_tokens, batch_context_tokens


class MentionPreprocessor:
    def __init__(self, tokenizer: PreTrainedTokenizerBase, length: int):
        self._tokenizer = tokenizer
        self._length = length
        self._windower = Windower(length)

    def __call__(
        self,
        batch_encoding: BatchEncoding,
        mention_indices: List[List[Tuple[int, int]]],
    ) -> Tuple[Mapping[str, torch.Tensor], int]:
        num_mentions = sum([len(doc_mentions) for doc_mentions in mention_indices])
        batch_span_tokens, batch_context_tokens = self._windower._get_spans_and_context(
            batch_encoding, mention_indices
        )
        encoded_mentions: Mapping[str, torch.Tensor] = self._tokenizer(
            batch_span_tokens,
            batch_context_tokens,
            truncation=True,
            padding="max_length",
            max_length=self._length,
            is_split_into_words=True,
            return_tensors="pt",
        )
        return encoded_mentions, num_mentions


class Preprocessor(NamedTuple):
    mention_preprocessor: MentionPreprocessor
    lkb_preprocessor: LKBPreprocessor

    def __call__(
        self,
        training_encodings: EncodedMentions,
        validation_encodings: EncodedMentions,
        lkb: LexicalKnowledgeBase,
    ) -> PreproccessedDataSet:
        candidate_encodings, lkb_index, num_candidates = self.lkb_preprocessor(lkb)

        return PreproccessedDataSet(
            self.get_dataset_partition(training_encodings, lkb_index),
            self.get_dataset_partition(validation_encodings, lkb_index),
            tensordict.TensorDict(
                dict(candidate_encodings), batch_size=[num_candidates]
            ),
        )

    #     return PreproccesedDatasetPartition(self.get_encodings(docs),self.get_concept_indices(docs,lkb_index))
    def get_dataset_partition(
        self, encoded_mentions: EncodedMentions, lkb_index: LKBIndex
    ) -> PreproccesedDatasetPartition:
        encoding_tensor, num_encodings = self.mention_preprocessor(
            encoded_mentions.doc_encodings, encoded_mentions.mention_indices
        )
        flattened_concept_ids = list(
            itertools.chain(*encoded_mentions.gold_concept_ids)
        )
        concept_indices = torch.tensor(
            [lkb_index[concept_id] for concept_id in flattened_concept_ids]
        )
        return PreproccesedDatasetPartition(
            tensordict.TensorDict(dict(encoding_tensor), batch_size=[num_encodings]),
            concept_indices,
        )


if __name__ == "__main__":
    pass
    # preprocessor.preprocess_medmentions(medmention_corpus)
