from abc import ABC, abstractmethod
import torch

from crtl.components.dual_encoder.pipeline.utils import convert_size


def get_dot_similarity(
    mention_embeddings: torch.Tensor, candidate_embeddings: torch.Tensor
) -> torch.Tensor:
    dot_similarity = torch.matmul(
        mention_embeddings.to(torch.device("cpu")),
        candidate_embeddings.to(torch.device("cpu")).T,
    )
    torch.mps.empty_cache()
    # cosine_similarity = dot_similarity / torch.norm(mention_embeddings,dim=1).unsqueeze(1).to(torch.device("cpu")) / torch.norm(candidate_embeddings,dim=1).unsqueeze(0).to(torch.device("cpu"))
    # return cosine_similarity.to(torch.device('cpu'))
    return dot_similarity


class Similarity(ABC):
    def __call__(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # (num_mentions x num_candidates_per_mention x embedding_dimension) OR (num_candidates x embedding_dimension)
    ) -> torch.Tensor:
        if len(candidate_embeddings.shape) == 3:
            return self.get_similarity_per_mention(
                mention_embeddings, candidate_embeddings
            )
        elif len(candidate_embeddings.shape) == 2:
            return self.get_cartesian_product_similarity(
                mention_embeddings, candidate_embeddings
            )
        else:
            raise ValueError(
                f"candidate_embeddings must have either 2 or 3 dimensions but has shape {candidate_embeddings.shape}"
            )

    @abstractmethod
    def get_similarity_per_mention(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_mentions x num_candidates_per_mention x embedding_dimension
    ) -> torch.Tensor:
        ...

    @abstractmethod
    def get_cartesian_product_similarity(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_candidates x embedding_dimension
    ) -> torch.Tensor:
        ...


class DotSimilarity(Similarity):
    def get_similarity_per_mention(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_mentions x num_candidates_per_mention x embedding_dimension
    ) -> torch.Tensor:
        return torch.bmm(
            candidate_embeddings,
            mention_embeddings.unsqueeze(-1),  # num_mentions x embeding_dimension x 1
        ).squeeze(
            -1
        )  # num_mentions x num_candidates_per_mention

    def get_cartesian_product_similarity(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_candidates x embedding_dimension
    ) -> torch.Tensor:
        dot_similarity = torch.matmul(
            mention_embeddings.to(torch.device("cpu")),
            candidate_embeddings.to(torch.device("cpu")).T,
        )
        torch.mps.empty_cache()
        # cosine_similarity = dot_similarity / torch.norm(mention_embeddings,dim=1).unsqueeze(1).to(torch.device("cpu")) / torch.norm(candidate_embeddings,dim=1).unsqueeze(0).to(torch.device("cpu"))
        # return cosine_similarity.to(torch.device('cpu'))
        return dot_similarity  # num_mentions x num_candidates


class CosineSimilarity:
    def get_similarity_per_mention(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_mentions x num_candidates_per_mention x embedding_dimension
    ) -> torch.Tensor:
        # mention_embeddings[:,None,:] : num_mentions x 1 x embedding_dimension
        return torch.cosine_similarity(
            mention_embeddings[:, None, :], candidate_embeddings, dim=-1
        )

    def get_cartesian_product_similarity(
        self,
        mention_embeddings: torch.Tensor,  # num_mentions x embedding_dimension
        candidate_embeddings: torch.Tensor,  # num_candidates x embedding_dimension
    ) -> torch.Tensor:
        return torch.cosine_similarity(
            mention_embeddings[None, :, :], candidate_embeddings[:, None, :], dim=-1
        )
        # alternatively:
        # dot_similarity = torch.matmul(
        # mention_embeddings.to(torch.device("cpu")),
        # candidate_embeddings.to(torch.device("cpu")).T,
        # )
        # torch.mps.empty_cache()
        # cosine_similarity = dot_similarity / torch.norm(mention_embeddings,dim=1).unsqueeze(1).to(torch.device("cpu")) / torch.norm(candidate_embeddings,dim=1).unsqueeze(0).to(torch.device("cpu"))
        # return cosine_similarity.to(torch.device('cpu'))
