import datetime
from tensordict import TensorDict, TensorDictBase

import torch
from crtl.components.dual_encoder.pipeline.loss import CrossEntropyScoreLoss
from crtl.components.dual_encoder.pipeline.linker_module import LinkerModule
from crtl.components.dual_encoder.pipeline.dataloader_creator import (
    DataLoaderCreator,
)
from crtl.components.dual_encoder.pipeline.embedder import EmbedderModule
from torch.utils.data import DataLoader
from torch.utils.tensorboard.writer import SummaryWriter
from crtl.components.dual_encoder.pipeline.preprocessor import PreproccessedDataSet
from crtl.components.dual_encoder.pipeline.score import Scorer
from crtl.components.dual_encoder.pipeline.similarity import get_dot_similarity
from crtl.components.dual_encoder.pipeline.utils import progressbar


class EntityLinkerTrainer:
    def __init__(
        self,
        mention_embedder_module: EmbedderModule,
        candidate_embedder_module: EmbedderModule,
        device: torch.device,
        batch_size: int = 30,
        num_epochs: int = 10,
        learning_rate: float = 1e-5,  # Recommended by https://arxiv.org/pdf/2103.05028.pdf
        verbose: bool = True,
        num_random_negatives: int = 5,
        num_hard_negatives: int = 5,
        adam_epsilon: float = 1e-8,
        warmup_steps: int = 0,
    ):
        """
        -------
        PARAMS:

        epoch_idx: int
            The number of the current epoch, starting at 0
        tb_writer: Optional[SummaryWriter]
            A SummaryWriter that batch-by-batch stats will be reported to. If None is provided, nothing will be reported.
        training_loader: DataLoader
            A DataLoader that contains the training data for the model. Each element should contain a triplet of a mention to encode, a postive concept text that is similar to it, and a negative concept text that is dissimilar to it
        optimizer: Optimizer
            The models' optimizer; usually set in the main train method
        loss_fn: _Loss
            The models' loss function; usually set in the main train method
        batch_size: int
            The size of each batch; summary stats will be computed and reported after each one has elapsed
        verbose: bool
            Whether to print summary stats to the console
        """
        self.linker_module = LinkerModule(
            mention_embedder_module, candidate_embedder_module
        ).to(device)
        self.loss_fn = CrossEntropyScoreLoss(batch_size)
        self.num_epochs = num_epochs
        self.learning_rate = learning_rate
        self._data_loader_getter = DataLoaderCreator(
            num_random_negatives, num_hard_negatives, device, batch_size
        )
        self.verbose = verbose
        self.start_time = datetime.datetime.now
        self.batch_size = batch_size
        self.writer = SummaryWriter(
            "runs/DE_CEL_{}".format(self.start_time().strftime("%Y%m%d_%H%M%S"))
        )
        self.num_epochs

        # AdamW optimizer recommended by https://arxiv.org/pdf/2103.05028.pdf
        self.optimizer = torch.optim.AdamW(
            self.linker_module.parameters(), lr=self.learning_rate, eps=adam_epsilon
        )
        self.scheduler = torch.optim.lr_scheduler.LinearLR(self.optimizer)
        self.batch_size = batch_size
        self.embedder = Embedder(batch_size=self.batch_size)
        self.scorer = Scorer()

    def train(self, dataset: "PreproccessedDataSet"):
        best_score = 0
        # training_mention_embeddings = self.embedder(dataset.training_dataset_partition.mention_encodings,self.linker_module.mention_embedder)
        candidate_embeddings = self.embedder(
            dataset.candidate_encodings, self.linker_module.candidate_embedder
        )
        score = self.eval(
            dataset.validation_dataset_partition.mention_encodings,
            dataset.validation_dataset_partition.gold_candidate_ids,
            candidate_embeddings,
        )
        self.writer.add_scalar("Score", score, 0)
        self.writer.flush()
        for epoch_idx in range(self.num_epochs):
            # Run training for one epoch and report back the avg. loss for the last batch
            # due to negative sampling, we have to create new data loaders every epoch
            if self.verbose:
                epoch_time = datetime.datetime.now()
                print("Now starting epoch {} at {}".format(epoch_idx, epoch_time))

            self.train_epoch(dataset, candidate_embeddings, epoch_idx)
            candidate_embeddings = self.embedder(
                dataset.candidate_encodings, self.linker_module.candidate_embedder
            )
            score = self.eval(
                dataset.validation_dataset_partition.mention_encodings,
                dataset.validation_dataset_partition.gold_candidate_ids,
                candidate_embeddings,
            )
            # avg_v_loss = self.get_new_avg_loss(eval_loader)
            self.writer.add_scalar("Score", score, epoch_idx + 1)
            self.writer.flush()
            # self.writer.add_scalars(
            #     "Training vs. Validation Loss",
            #     {"Training": avg_training_loss, "Validation": avg_v_loss},
            #     epoch_idx,
            # )
            # Save the model if it's beat its own record
            if score > best_score:
                best_score = score
                # best_v_loss = avg_v_loss
                model_path = "models/model_{}_{}".format(self.start_time, epoch_idx)
                torch.save(self.linker_module, "best_model.pth")
        if self.verbose:
            print("Done! At {}".format(datetime.datetime.now()))
        torch.save(self.linker_module, "last_model.pth")

    def eval(
        self,
        mention_encodings: TensorDictBase,
        gold_candidate_ids: torch.Tensor,
        candidate_embeddings: torch.Tensor,
    ) -> float:
        validation_mention_embeddings = self.embedder(
            mention_encodings, self.linker_module.mention_embedder
        )
        self.linker_module.eval()
        with torch.no_grad():
            cosine_similarity = get_dot_similarity(
                validation_mention_embeddings, candidate_embeddings
            )
            return self.scorer.recall_at_n_score(
                gold_candidate_ids, similarity_scores=cosine_similarity, n=10
            )

    def train_epoch(self, dataset, candidate_embeddings, epoch_idx: int) -> float:
        """
        train_epoch

        Trains the underlying models in the encoders for one epoch. Returns the average loss for the last batch

        """
        training_data_loader = self.get_training_dataloader(
            dataset, candidate_embeddings
        )
        running_loss = 0.0
        print("Training...")
        for idx, batch_data in enumerate(
            progressbar(training_data_loader, len(training_data_loader))
        ):
            # Type: Tuple[str, torch.Tensor, List[torch.Tensor]]
            loss_value = self.train_batch(batch_data)
            running_loss += loss_value
            if self.verbose and idx % 10 == 0 and idx != 0:
                loss_per_batch = running_loss / 10
                # print("\tbatch {} loss: {}".format(idx + 1, loss_per_batch))
                tb_idx = epoch_idx * len(training_data_loader) + idx
                if self.writer is not None:
                    self.writer.add_scalar("Loss/train", loss_per_batch, tb_idx)
                    self.writer.flush()
                running_loss = 0
        return loss_per_batch

    def get_training_dataloader(self, dataset, candidate_embeddings):
        training_mention_embeddings = self.embedder(
            dataset.training_dataset_partition.mention_encodings,
            self.linker_module.mention_embedder,
        )
        with torch.no_grad():
            training_loader = self._data_loader_getter(
                dataset.training_dataset_partition.gold_candidate_ids,
                dataset.training_dataset_partition.mention_encodings,
                dataset.candidate_encodings,
                training_mention_embeddings,
                candidate_embeddings,
            )
            return training_loader

    def train_batch(self, x: TensorDict):
        self.optimizer.zero_grad()
        (
            mention_embedding,
            positive_embedding,
            negative_embeddings,
        ) = self.linker_module.forward(x)
        loss_value = self.loss_fn(
            mention_embedding, positive_embedding, negative_embeddings
        )
        loss_value.backward()

        self.optimizer.step()
        return loss_value.detach().item()

    def get_new_avg_loss(self, eval_data_loader: DataLoader[TensorDictBase]):
        with torch.no_grad():
            v_data_size = 0
            running_v_loss = 0
            for features in eval_data_loader:
                (
                    mention_embedding,
                    positive_embedding,
                    negative_embeddings,
                ) = self.linker_module.forward(features)
                running_v_loss += self.loss_fn.forward(
                    mention_embedding,
                    positive_embedding,
                    negative_embeddings,
                )
                v_data_size += 1

            avg_v_loss = running_v_loss / (v_data_size + 1)
        return avg_v_loss


class Embedder:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size

    def __call__(self, encodings: TensorDictBase, embedder_module: EmbedderModule):
        return self.embed(self.get_dataloader(encodings), embedder_module)

    def get_dataloader(self, encodings: TensorDictBase) -> DataLoader[TensorDictBase]:
        # embedding_dataset = EncodingDataset(encodings)
        return DataLoader(encodings, batch_size=self.batch_size, collate_fn=lambda x: x)

    @staticmethod
    def embed(
        embedding_dataloader: DataLoader[TensorDictBase],
        embedder_module: EmbedderModule,
    ) -> torch.Tensor:
        embeddings = []
        with torch.no_grad():
            for encodings in progressbar(
                embedding_dataloader, len(embedding_dataloader)
            ):
                # print(encodings)
                # encodings = encodings.to(mps_device)
                # print("device:",encodings.device)
                embeddings.append(embedder_module.forward(encodings))

            return torch.cat(embeddings)
